"""A collection of CLI scripts."""
import argparse
import codecs
import json
import sys
from logging import basicConfig

from pyknp import KNP

from pyknp_eventgraph import EventGraph, make_image


def evg():
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", "-o", default="", help="path to output")
    args = parser.parse_args()

    basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    knp = KNP()
    results = []
    chunk = ""
    for line in codecs.getreader("utf-8")(getattr(sys.stdin, "buffer", sys.stdin)):
        chunk += line
        if line.strip() == "EOS":
            results.append(knp.result(chunk))
            chunk = ""
    evg_ = EventGraph.build(results)
    if args.output:
        evg_.save(args.output)
    else:
        print(json.dumps(evg_.to_dict(), indent=4, ensure_ascii=False))


def evgviz():
    parser = argparse.ArgumentParser()
    parser.add_argument("IN", help="path to input")
    parser.add_argument("OUT", help="path to output")
    parser.add_argument("--exclude-detail", action="store_true", help="exclude detail information of events")
    parser.add_argument("--exclude-original-text", action="store_true", help="exclude original texts")
    parser.add_argument("--binary", "-b", action="store_true", help="whether the input is binary")
    args = parser.parse_args()

    if args.binary:
        with open(args.IN, "rb") as f:
            evg_ = EventGraph.load(f, binary=args.binary)
    else:
        with open(args.IN) as f:
            evg_ = EventGraph.load(f, binary=args.binary)

    make_image(evg_, args.OUT, with_detail=not args.exclude_detail, with_original_text=not args.exclude_original_text)



from pyknp_eventgraph.argument import Argument
from pyknp_eventgraph.base_phrase import BasePhrase
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.document import Document
from pyknp_eventgraph.event import Event
from pyknp_eventgraph.eventgraph import EventGraph
from pyknp_eventgraph.features import Features
from pyknp_eventgraph.pas import PAS
from pyknp_eventgraph.predicate import Predicate
from pyknp_eventgraph.relation import Relation
from pyknp_eventgraph.sentence import Sentence
from pyknp_eventgraph.visualizer import make_image



"""The base class of EventGraph components."""
from abc import ABC, abstractmethod


class Component(ABC):
    """The base of EventGraph components."""

    def __repr__(self) -> str:
        return self.to_string()

    @abstractmethod
    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        raise NotImplementedError

    @abstractmethod
    def to_string(self) -> str:
        """Convert this object into a string."""
        raise NotImplementedError



from typing import List

from pyknp import Tag

PAS_ORDER = {"ガ２": 0, "ガ": 1, "ヲ": 2, "ニ": 3}


def get_parallel_tags(tag: Tag) -> List[Tag]:
    """Return parallel tags of a given tag.

    Args:
        tag: A :class:`pyknp.knp.tag.Tag` object.

    Returns:
        A list of parallel tags.
    """
    parallels = []
    while tag.dpndtype == "P":
        parallels.append(tag.parent)
        tag = tag.parent
    return parallels


def convert_katakana_to_hiragana(in_str: str) -> str:
    """Convert katakana characters in a given string to their corresponding hiragana characters.

    Args:
        in_str: A string.

    Returns:
        A string where katakana characters have been converted into hiragana.
    """
    return "".join(chr(ord(ch) - 96) if ("ァ" <= ch <= "ン") else ch for ch in in_str)


def convert_mrphs_to_surf(mrphs: str) -> str:
    """Remove unnecessary spaces from a tokenized surface string."""
    surf = mrphs.replace(" ", "")
    surf = surf.replace("]", "] ").replace("|", " | ").replace("▼", "▼ ").replace("■", "■ ").replace("(", " (")
    return surf.strip()



import json
import pickle
from logging import getLogger
from typing import BinaryIO, List, Optional, TextIO, Union

from pyknp import BList

from pyknp_eventgraph.base_phrase import BasePhraseBuilder
from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.document import Document, DocumentBuilder, JsonDocumentBuilder
from pyknp_eventgraph.event import Event
from pyknp_eventgraph.relation import JsonRelationBuilder, Relation, RelationsBuilder
from pyknp_eventgraph.sentence import Sentence

logger = getLogger(__name__)


class EventGraph(Component):
    """EventGraph provides a high-level interface that facilitates NLP application development. The core concept of
    EventGraph is event, a language information unit that is closely related to predicate-argument structure but more
    application-oriented. Events are linked to each other based on their syntactic and semantic relations.

    Attributes:
        document (Document): A document on which this EventGraph is built.
    """

    def __init__(self):
        self.document: Optional[Document] = None

    @classmethod
    def build(cls, blist: List[BList]) -> "EventGraph":
        """Build an EventGraph from language analysis by KNP.

        Args:
            blist: A list of bunsetsu lists, each of which is a result of analysis performed by KNP on a sentence.

        Example::

            from pyknp import KNP
            from pyknp_eventgraph import EventGraph

            # Parse a document.
            document = ['彼女は海外勤務が長いので、英語がうまいに違いない。', '私はそう確信していた。']
            knp = KNP()
            blists = [knp.parse(sentence) for sentence in document]

            # Build an EventGraph.
            evg = EventGraph.build(blists)
        """
        return EventGraphBuilder()(blist)

    @classmethod
    def load(cls, f: Union[TextIO, BinaryIO], binary: bool = False) -> "EventGraph":
        """Deserialize an EventGraph.

        Args:
            f: A file descriptor.
            binary: If true, deserialize an EventGraph using Python's pickle utility. Otherwise, deserialize
                an EventGraph using Python's json utility.

        Example::

            from pyknp_eventgraph import EventGraph

            # Load an EventGraph serialized in a JSON format.
            with open('evg.json', 'r') as f:
                evg = EventGraph.load(f, binary=False)

            # Load an EventGraph serialized by Python's pickle utility.
            with open('evg.pkl', 'rb') as f:
                evg = EventGraph.load(f, binary=True)

        Caution:
            EventGraph deserialized from a JSON file loses several functionality.
            To keep full functionality, use Python\'s pickle utility for serialization.
        """
        return PickleEventGraphBuilder()(f) if binary else JsonEventGraphBuilder()(f)

    def save(self, path: str, binary: bool = False) -> None:
        """Save this EventGraph.

        Args:
            path: An output file path.
            binary: If true, serialize this EventGraph using Python's pickle utility. Otherwise, serialize
                this EventGraph using Python's json utility.

        Caution:
            EventGraph deserialized from a JSON file loses several functionality. To keep full functionality,
            use Python\'s pickle utility for serialization.
        """
        if binary:
            with open(path, "wb") as f:
                pickle.dump(self, f)
        else:
            logger.info(
                "EventGraph deserialized from a JSON file loses several functionality. "
                "To keep full functionality, use Python's pickle utility for serialization. "
                "For details, refer to https://pyknp-eventgraph.readthedocs.io/en/latest/reference/"
                "eventgraph.html#pyknp_eventgraph.eventgraph.EventGraph.save."
            )
            with open(path, "w") as f:
                json.dump(self.to_dict(), f, ensure_ascii=False, indent=8)

    @property
    def sentences(self) -> List[Sentence]:
        """A list of sentences."""
        return [sentence for sentence in self.document.sentences]

    @property
    def events(self) -> List[Event]:
        """A list of events."""
        return [event for sentence in self.sentences for event in sentence.events]

    @property
    def relations(self) -> List[Relation]:
        """A list of relations."""
        return [relation for event in self.events for relation in event.outgoing_relations]

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            sentences=[sentence.to_dict() for sentence in self.sentences],
            events=[event.to_dict() for event in self.events],
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return (
            f"<EventGraph, "
            f"#sentences: {len(self.sentences)}, "
            f"#events: {len(self.events)}, "
            f"#relations: {len(self.relations)}>"
        )


class EventGraphBuilder(Builder):
    def __call__(self, blists: List[BList]) -> EventGraph:
        logger.debug("Create an EventGraph.")
        Builder.reset()
        evg = EventGraph()
        # Assign a document to the EventGraph.
        # A document is a collection of sentences, and a sentence is a collection of events.
        DocumentBuilder()(evg, blists)
        # Assign basic phrases to events.
        # This process must be performed after constructing a document
        # because an event may have a basic phrase recognized by inter-sentential cataphora resolution.
        for event in evg.events:
            BasePhraseBuilder()(event)
        # Assign event-to-event relations to events.
        # This process must be performed after constructing a document.
        for event in evg.events:
            RelationsBuilder()(event)
        logger.debug("Successfully created an EventGraph.")
        logger.debug(evg)
        return evg


class PickleEventGraphBuilder(Builder):
    def __call__(self, f: BinaryIO) -> EventGraph:
        logger.debug("Create an EventGraph by loading a pickled file.")
        evg = pickle.load(f)
        assert isinstance(evg, EventGraph)
        logger.debug("Successfully created an EventGraph.")
        logger.debug(evg)
        return evg


class JsonEventGraphBuilder(Builder):
    def __call__(self, f: TextIO) -> EventGraph:
        logger.debug("Create an EventGraph by loading a JSON file.")
        logger.info(
            "EventGraph deserialized from a JSON file loses several functionality. "
            "To keep full functionality, use Python's pickle utility for serialization. "
            "For details, refer to https://pyknp-eventgraph.readthedocs.io/en/latest/reference/eventgraph.html"
            "#pyknp_eventgraph.eventgraph.EventGraph.load."
        )
        Builder.reset()
        dump = json.load(f)
        evg = EventGraph()
        # Assign a document to the EventGraph.
        # A document is a collection of sentences, and a sentence is a collection of events.
        JsonDocumentBuilder()(evg, dump)
        # Assign event-to-event relations to events.
        for event_dump in dump["events"]:
            for relation_dump in event_dump["rel"]:
                modifier_evid = event_dump["event_id"]
                head_evid = relation_dump["event_id"]
                JsonRelationBuilder()(modifier_evid, head_evid, relation_dump)
        logger.debug("Successfully created an EventGraph.")
        logger.debug(evg)
        return evg



import re
from logging import getLogger
from typing import TYPE_CHECKING, List, Optional

from pyknp import Tag

from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event

logger = getLogger(__name__)


class Features(Component):
    """Features provides linguistic information of an event.

    Attributes:
        event (Event): An event.
        modality (List[str]): A list of modality, a linguistic expression that indicates how a write judges and feels
            about content. Each of item can take either "意志 (volition)," "勧誘 (invitation)," "命令 (imperative),"
            "禁止 (prohibition)," "評価:弱 (evaluation: weak)," "評価:強 (evaluation: strong),"
            "認識-推量 (certainty-subjective)," "認識-蓋然性 (certainty-epistemic)," "認識-証拠 (certainty-evidential),"
            "依頼Ａ (request-A)," "依頼Ｂ (request-B)," and "推量・伝聞 (supposition/hearsay)."
        tense (str): The place of an event in a time frame, which can take either "過去 (past)" or "非過去 (non-past)."
        negation (bool): If true, this event uses a negative construction.
        state (str): A type of a predicate, which can take either "動態述語 (action)" or "状態述語 (state)."
        complement (bool): If true, this event modifies an event as a sentential complementizer.
        level (str, optional): The semantic heaviness of a predicate.
    """

    def __init__(
        self,
        event: "Event",
        modality: List[str],
        tense: str,
        negation: bool,
        state: str,
        complement: bool,
        level: Optional[str] = None,
    ):
        self.event: Event = event
        self.modality: List[str] = modality
        self.tense: str = tense
        self.negation: bool = negation
        self.state: str = state
        self.complement: bool = complement
        self.level: Optional[str] = level

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            modality=self.modality,
            tense=self.tense,
            negation=self.negation,
            state=self.state,
            complement=self.complement,
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return (
            f"<Features, "
            f'modality: {", ".join(self.modality) if self.modality else "None"}, '
            f"tense: {self.tense}, "
            f"negation: {self.negation}, "
            f"state: {self.state}, "
            f"complement: {self.complement}>"
        )


class FeaturesBuilder(Builder):
    def __call__(self, event: "Event") -> Features:
        func_tag = self._get_functional_tag(event.head)
        features = Features(
            event=event,
            modality=self._find_modality(event.head, func_tag),
            tense=self._find_tense(func_tag),
            negation=self._find_negation(func_tag),
            state=self._find_state(func_tag),
            complement=self._find_complement(func_tag),
            level=self._find_level(func_tag),
        )
        event.features = features
        return features

    @staticmethod
    def _get_functional_tag(head: Tag) -> Tag:
        if (
            head.parent
            and head.parent.pas
            and "用言" in head.parent.features
            and "修飾" not in head.parent.features
            and "機能的基本句" in head.parent.features
        ):
            return head.parent
        return head

    @staticmethod
    def _find_modality(head: Tag, func_tag: Tag) -> List[str]:
        modality = re.findall("<モダリティ-(.+?)>", func_tag.fstring)
        if head.parent and ("弱用言" in head.parent.features or "思う能動" in head.parent.features):
            modality.append("推量・伝聞")
        return modality

    @staticmethod
    def _find_tense(func_tag: Tag) -> str:
        if "<時制" in func_tag.fstring:
            return re.search("<時制[-:](.+?)>", func_tag.fstring).group(1)
        return "unknown"

    @staticmethod
    def _find_negation(func_tag: Tag) -> bool:
        return func_tag.features.get("否定表現", False)

    @staticmethod
    def _find_state(head: Tag) -> str:
        if "状態述語" in head.features:
            return "状態述語"
        if "動態述語" in head.features:
            return "動態述語"
        return ""

    @staticmethod
    def _find_complement(func_tag: Tag) -> bool:
        return func_tag.features.get("補文", False)

    @staticmethod
    def _find_level(func_tag: Tag) -> str:
        return func_tag.features.get("レベル", "")


class JsonFeaturesBuilder(Builder):
    def __call__(self, event: "Event", dump: dict) -> Features:
        features = Features(
            event=event,
            modality=dump["modality"],
            tense=dump["tense"],
            negation=dump["negation"],
            state=dump["state"],
            complement=dump["complement"],
        )
        event.features = features
        return features



from logging import getLogger
from typing import TYPE_CHECKING, List, Optional

from pyknp import Morpheme, Tag

from pyknp_eventgraph.base_phrase import BasePhrase
from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.helper import convert_mrphs_to_surf

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event
    from pyknp_eventgraph.pas import PAS

logger = getLogger(__name__)


class Predicate(Component):
    """A predicate is the core of a　PAS.

    Attributes:
        pas (PAS): A PAS that this predicate belongs.
        head (:class:`pyknp.knp.tag.Tag`): A head tag.
        type_ (str): A type of this predicate.
        head_base_phrase (Token, optional): A head basic phrase.
    """

    def __init__(self, pas: "PAS", type_: str, head: Optional[Tag] = None):
        self.pas: PAS = pas
        self.type_: str = type_
        self.head: Optional[Tag] = head
        self.head_base_phrase: Optional[BasePhrase] = None

        self._surf = None
        self._normalized_surf = None
        self._mrphs = None
        self._normalized_mrphs = None
        self._reps = None
        self._normalized_reps = None
        self._standard_reps = None
        self._children = None
        self._adnominal_event_ids = None
        self._sentential_complement_event_ids = None

    @property
    def tag(self) -> Optional[Tag]:
        """The tag of the head base phrase."""
        return self.head_base_phrase.tag

    @property
    def surf(self) -> str:
        """A surface string."""
        if self._surf is None:
            self._surf = convert_mrphs_to_surf(self.mrphs)
        return self._surf

    @property
    def normalized_surf(self) -> str:
        """A normalized surface string."""
        return self.surf

    @property
    def mrphs(self) -> str:
        """A tokenized string."""
        if self._mrphs is None:
            mrphs = []
            is_within_standard_repname = False
            for bp in self.head_base_phrase.modifiees(include_self=True):
                for m in bp.tag.mrph_list():
                    if "用言表記先頭" in m.fstring:
                        is_within_standard_repname = True
                    if "用言表記末尾" in m.fstring:
                        mrphs.append(m.genkei)  # Normalize the last morpheme.
                        return " ".join(mrphs)
                    if is_within_standard_repname:
                        mrphs.append(m.midasi)
            self._mrphs = " ".join(mrphs)
        return self._mrphs

    @property
    def normalized_mrphs(self) -> str:
        """A tokenized/normalized surface string."""
        return self.mrphs

    @property
    def reps(self) -> str:
        """A representative string."""
        if self._reps is None:
            for bp in self.head_base_phrase.modifiees(include_self=True):
                if "用言代表表記" in bp.tag.features:
                    self._reps = bp.tag.features["用言代表表記"]
                    break
            else:
                self._reps = self._base_phrase_to_text(
                    self.head_base_phrase, mode="reps", truncate=True, include_modifiees=True
                )
        return self._reps

    @property
    def normalized_reps(self) -> str:
        """A normalized representative string."""
        return self.reps

    @property
    def standard_reps(self) -> str:
        """A standard representative string."""
        if self._standard_reps is None:
            for bp in self.head_base_phrase.modifiees(include_self=True):
                if "標準用言代表表記" in bp.tag.features:
                    self._standard_reps = bp.tag.features["標準用言代表表記"]
                    break
            else:
                self._standard_reps = self.reps
        return self._standard_reps

    @property
    def type(self) -> str:
        """The type of this predicate."""
        return self.type_

    @property
    def adnominal_events(self) -> List["Event"]:
        """A list of events modifying this predicate as an adnominal."""
        return [e for bp in self.head_base_phrase.modifiees(include_self=True) for e in bp.adnominal_events]

    @property
    def sentential_complement_events(self) -> List["Event"]:
        """A list of events modifying this predicate as an adnominal."""
        return [e for bp in self.head_base_phrase.modifiees(include_self=True) for e in bp.sentential_complement_events]

    @property
    def adnominal_event_ids(self) -> List[int]:
        """A list of IDs of events modifying this predicate (adnominal)."""
        if self._adnominal_event_ids is None:
            self._adnominal_event_ids = sorted(e.evid for e in self.adnominal_events)
        return self._adnominal_event_ids

    @property
    def sentential_complement_event_ids(self) -> List[int]:
        """A list of IDs of events modifying this predicate (sentential complement)."""
        if self._sentential_complement_event_ids is None:
            self._sentential_complement_event_ids = sorted(e.evid for e in self.sentential_complement_events)
        return self._sentential_complement_event_ids

    @property
    def children(self) -> List[dict]:
        """A list of child words."""
        if self._children is None:
            self._children = []
            for bp in reversed(self.head_base_phrase.modifiers()):
                self._children.append(
                    {
                        "surf": convert_mrphs_to_surf(self._base_phrase_to_text(bp, mode="mrphs", truncate=False)),
                        "normalized_surf": convert_mrphs_to_surf(
                            self._base_phrase_to_text(bp, mode="mrphs", truncate=True)
                        ),
                        "mrphs": self._base_phrase_to_text(bp, mode="mrphs", truncate=False),
                        "normalized_mrphs": self._base_phrase_to_text(bp, mode="mrphs", truncate=True),
                        "reps": self._base_phrase_to_text(bp, mode="reps", truncate=False),
                        "normalized_reps": self._base_phrase_to_text(bp, mode="reps", truncate=True),
                        "adnominal_event_ids": [event.evid for event in bp.adnominal_events],
                        "sentential_complement_event_ids": [event.evid for event in bp.sentential_complement_events],
                        "modifier": "修飾" in bp.tag.features,
                        "possessive": bp.tag.features.get("係", "") == "ノ格",
                    }
                )
        return self._children

    def _base_phrase_to_text(
        self, bp: BasePhrase, mode: str = "mrphs", truncate: bool = False, include_modifiees: bool = False
    ) -> str:
        """Convert a base phrase to a text.

        Args:
            bp: A base phrase.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            truncate: If true, adjunct words are truncated.
            include_modifiees: If true, parents are used to construct a compound phrase.
        """
        assert mode in {"mrphs", "reps"}
        mrphs = list(bp.tag.mrph_list())
        if include_modifiees:
            for parent_bp in bp.modifiees():
                mrphs += list(parent_bp.tag.mrph_list())
        if truncate:
            mrphs = self._truncate_mrphs(mrphs)
            return self._format_mrphs(mrphs, mode, normalize=True)
        else:
            return self._format_mrphs(mrphs, mode, normalize=False)

    @staticmethod
    def _truncate_mrphs(mrphs: List[Morpheme]) -> List[Morpheme]:
        """Truncate a list of morphemes.

        Args:
            mrphs: A list of morphemes.

        """
        for index, mrph in reversed(list(enumerate(mrphs))):
            if mrph.hinsi == "助動詞" and mrph.genkei == "です" and 0 < index and mrphs[index - 1].hinsi == "形容詞":
                # adjective + 'です' -> ignore 'です' (e.g., 美しいです -> 美しい)
                return mrphs[:index]
            elif mrph.hinsi == "判定詞" and mrph.midasi == "じゃ" and 0 < index and "<活用語>" in mrphs[index - 1].fstring:
                # adjective or verb +'じゃん' -> ignore 'じゃん' (e.g., 使えないじゃん -> 使えない)
                return mrphs[:index]
            elif ("<活用語>" in mrph.fstring or "<用言意味表記末尾>" in mrph.fstring) and mrph.genkei not in {"のだ", "んだ"}:
                # check the last word with conjugation except some meaningless words
                return mrphs[: index + 1]
        return mrphs

    @staticmethod
    def _format_mrphs(mrphs: List[Morpheme], mode: str, normalize: bool = False) -> str:
        """Convert a list of morphemes to a text.

        Args:
            mrphs: A list of morphemes.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            normalize: If true, the last content word will be normalized.

        """
        assert mode in {"mrphs", "reps"}
        if mode == "reps":
            return " ".join(mrph.repname or f"{mrph.midasi}/{mrph.midasi}" for mrph in mrphs)
        else:  # i.e., mode == 'mrphs'
            if normalize:
                # Change the last morpheme to its infinitive (i.e., genkei)
                base = " ".join(mrph.midasi for mrph in mrphs[:-1])
                if mrphs[-1].hinsi == "助動詞" and mrphs[-1].genkei == "ぬ":
                    # Exception to prevent transforming "できません" into "できませぬ".
                    return f"{base} {mrphs[-1].midasi}".strip()
                else:
                    return f"{base} {mrphs[-1].genkei}".strip()
            else:
                return " ".join(mrph.midasi for mrph in mrphs)

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            surf=self.surf,
            normalized_surf=self.normalized_surf,
            mrphs=self.mrphs,
            normalized_mrphs=self.normalized_mrphs,
            reps=self.reps,
            normalized_reps=self.normalized_reps,
            standard_reps=self.standard_reps,
            type=self.type,
            adnominal_event_ids=self.adnominal_event_ids,
            sentential_complement_event_ids=self.sentential_complement_event_ids,
            children=self.children,
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Predicate, type: {self.type_}, surf: {self.surf}>"


class PredicateBuilder(Builder):
    def __call__(self, pas: "PAS") -> Predicate:
        predicate = Predicate(pas, self._find_type(pas.event.head), pas.event.head)
        pas.predicate = predicate
        return predicate

    @staticmethod
    def _find_type(head: Tag) -> str:
        return head.features.get("用言", "")


class JsonPredicateBuilder(Builder):
    def __call__(self, pas: "PAS", dump: dict) -> Predicate:
        predicate = Predicate(pas, dump["type"])
        predicate._surf = dump["surf"]
        predicate._normalized_surf = dump["normalized_surf"]
        predicate._mrphs = dump["mrphs"]
        predicate._normalized_mrphs = dump["normalized_mrphs"]
        predicate._reps = dump["reps"]
        predicate._normalized_reps = dump["normalized_reps"]
        predicate._standard_reps = dump["standard_reps"]
        predicate._children = dump["children"]
        predicate._adnominal_event_ids = dump["adnominal_event_ids"]
        predicate._sentential_complement_event_ids = dump["sentential_complement_event_ids"]
        pas.predicate = predicate
        return predicate



from logging import getLogger
from typing import TYPE_CHECKING, List, Optional

from pyknp import BList, Tag

from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.event import Event, EventBuilder
from pyknp_eventgraph.helper import convert_mrphs_to_surf

if TYPE_CHECKING:
    from pyknp_eventgraph.document import Document

logger = getLogger(__name__)


class Sentence(Component):
    """A sentence is a collection of events.

    Attributes:
        document (Document): A document that includes this sentence.
        sid (str): An original sentence ID.
        ssid (int): A serial sentence ID.
        blist (:class:`pyknp.knp.blist.BList`, optional): A list of bunsetsu-s.
        events (List[Event]): A list of events in this sentence.
    """

    def __init__(self, document: "Document", sid: str, ssid: int, blist: Optional[BList] = None):
        self.document: Document = document
        self.sid: str = sid
        self.ssid: int = ssid
        self.blist: BList = blist
        self.events: List[Event] = []

        self._mrphs = None
        self._reps = None

    @property
    def surf(self) -> str:
        """A surface string."""
        return convert_mrphs_to_surf(self.mrphs)

    @property
    def mrphs(self) -> str:
        """A tokenized surface string."""
        if self._mrphs is None:
            self._mrphs = " ".join(m.midasi for m in self.blist.mrph_list())
        return self._mrphs

    @property
    def reps(self) -> str:
        """A representative string."""
        if self._reps is None:
            self._reps = " ".join(m.repname or f"{m.midasi}/{m.midasi}" for m in self.blist.mrph_list())
        return self._reps

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(sid=self.sid, ssid=self.ssid, surf=self.surf, mrphs=self.mrphs, reps=self.reps)

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Sentence, sid: {self.sid}, ssid: {self.ssid}, surf: {self.surf}>"


class SentenceBuilder(Builder):
    def __call__(self, document: "Document", blist: BList) -> Sentence:
        sentence = Sentence(document, blist.sid, Builder.ssid, blist)
        start: Optional[Tag] = None
        end: Optional[Tag] = None
        head: Optional[Tag] = None
        for tag in blist.tag_list():
            if not start:
                start = tag
            if not head and "節-主辞" in tag.features:
                head = tag
            if not end and "節-区切" in tag.features:
                end = tag
                if head:
                    EventBuilder()(sentence, start, head, end)
                start, end, head = None, None, None
        document.sentences.append(sentence)
        Builder.ssid += 1
        for bid, bnst in enumerate(blist.bnst_list()):
            for tag in bnst.tag_list():
                Builder.stid_bid_map[(sentence.ssid, tag.tag_id)] = bid
                Builder.stid_tag_map[(sentence.ssid, tag.tag_id)] = tag
        return sentence


class JsonSentenceBuilder(Builder):
    def __call__(self, document: "Document", dump: dict) -> Sentence:
        sentence = Sentence(document, dump["sid"], dump["ssid"])
        sentence._mrphs = dump["mrphs"]
        sentence._reps = dump["reps"]
        document.sentences.append(sentence)
        Builder.ssid += 1
        return sentence



import collections
from logging import getLogger
from typing import TYPE_CHECKING, Dict, List, Optional

from pyknp import Argument as PyknpArgument
from pyknp import Morpheme, Tag

from pyknp_eventgraph.base_phrase import BasePhrase
from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.helper import PAS_ORDER, convert_katakana_to_hiragana, convert_mrphs_to_surf

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event
    from pyknp_eventgraph.pas import PAS

logger = getLogger(__name__)


class Argument(Component):
    """An argument supplements its predicate's information.

    Attributes:
        pas (PAS): A PAS that this argument belongs.
        case (str): A case.
        eid (int): An entity ID.
        flag (str): A flag.
        sdist (int): The sentence distance between this argument and the predicate.
        arg (:class:`pyknp.knp.pas.Argument`, optional): An Argument object in pyknp.
        head_base_phrase (BasePhrase, optional): A head basic phrase.
    """

    def __init__(self, pas: "PAS", case: str, eid: int, flag: str, sdist: int, arg: Optional[PyknpArgument] = None):
        self.pas: "PAS" = pas
        self.case: str = case
        self.eid: int = eid
        self.flag: str = flag
        self.sdist: int = sdist
        self.arg: Optional[PyknpArgument] = arg
        self.head_base_phrase: Optional[BasePhrase] = None

        self._surf = None
        self._normalized_surf = None
        self._mrphs = None
        self._normalized_mrphs = None
        self._reps = None
        self._normalized_reps = None
        self._head_reps = None
        self._children = None
        self._adnominal_event_ids = None
        self._sentential_complement_event_ids = None

    @property
    def tag(self) -> Optional[Tag]:
        """The tag of the head base phrase."""
        return self.head_base_phrase.tag

    @property
    def surf(self) -> str:
        """A surface string."""
        if self._surf is None:
            self._surf = convert_mrphs_to_surf(self.mrphs)
        return self._surf

    @property
    def normalized_surf(self) -> str:
        """A normalized surface string."""
        if self._normalized_surf is None:
            self._normalized_surf = convert_mrphs_to_surf(self.normalized_mrphs)
        return self._normalized_surf

    @property
    def mrphs(self) -> str:
        """A tokenized surface string."""
        if self._mrphs is None:
            self._mrphs = self._base_phrase_to_text(self.head_base_phrase, truncate=False, include_modifiees=True)
        return self._mrphs

    @property
    def normalized_mrphs(self) -> str:
        """A tokenized/normalized surface string."""
        if self._normalized_mrphs is None:
            self._normalized_mrphs = self._base_phrase_to_text(
                self.head_base_phrase, truncate=True, include_modifiees=True
            )
        return self._normalized_mrphs

    @property
    def reps(self) -> str:
        """A representative string."""
        if self._reps is None:
            self._reps = self._base_phrase_to_text(
                self.head_base_phrase, mode="reps", truncate=False, include_modifiees=True
            )
        return self._reps

    @property
    def normalized_reps(self) -> str:
        """A normalized representative string."""
        if self._normalized_reps is None:
            self._normalized_reps = self._base_phrase_to_text(
                self.head_base_phrase, mode="reps", truncate=True, include_modifiees=True
            )
        return self._normalized_reps

    @property
    def head_reps(self) -> str:
        """A head representative string."""
        if self._head_reps is None:
            if self.head_base_phrase.tag:  # Not an exophora.
                head_reps = self.head_base_phrase.tag.head_prime_repname or self.head_base_phrase.tag.head_repname
                if head_reps:
                    self._head_reps = f"[{head_reps}]" if self.head_base_phrase.omitted_case else head_reps
            self._head_reps = self._head_reps or self.normalized_reps
        return self._head_reps

    @property
    def adnominal_events(self) -> List["Event"]:
        """A list of events modifying this predicate as an adnominal."""
        return [e for bp in self.head_base_phrase.modifiees(include_self=True) for e in bp.adnominal_events]

    @property
    def sentential_complement_events(self) -> List["Event"]:
        """A list of events modifying this predicate as an adnominal."""
        return [e for bp in self.head_base_phrase.modifiees(include_self=True) for e in bp.sentential_complement_events]

    @property
    def adnominal_event_ids(self) -> List[int]:
        """A list of IDs of events modifying this predicate (adnominal)."""
        if self._adnominal_event_ids is None:
            self._adnominal_event_ids = sorted(e.evid for e in self.adnominal_events)
        return self._adnominal_event_ids

    @property
    def sentential_complement_event_ids(self) -> List[int]:
        """A list of IDs of events modifying this predicate (sentential complement)."""
        if self._sentential_complement_event_ids is None:
            self._sentential_complement_event_ids = sorted(e.evid for e in self.sentential_complement_events)
        return self._sentential_complement_event_ids

    @property
    def children(self) -> List[dict]:
        """A list of child words."""
        if self._children is None:
            self._children = []
            for bp in reversed(self.head_base_phrase.modifiers()):
                self._children.append(
                    {
                        "surf": convert_mrphs_to_surf(self._base_phrase_to_text(bp, mode="mrphs", truncate=False)),
                        "normalized_surf": convert_mrphs_to_surf(
                            self._base_phrase_to_text(bp, mode="mrphs", truncate=True)
                        ),
                        "mrphs": self._base_phrase_to_text(bp, mode="mrphs", truncate=False),
                        "normalized_mrphs": self._base_phrase_to_text(bp, mode="mrphs", truncate=True),
                        "reps": self._base_phrase_to_text(bp, mode="reps", truncate=False),
                        "normalized_reps": self._base_phrase_to_text(bp, mode="reps", truncate=True),
                        "adnominal_event_ids": [e.evid for e in bp.adnominal_events],
                        "sentential_complement_event_ids": [e.evid for e in bp.sentential_complement_events],
                        "modifier": "修飾" in bp.tag.features,
                        "possessive": bp.tag.features.get("係", "") == "ノ格",
                    }
                )
        return self._children

    def _base_phrase_to_text(
        self, bp: BasePhrase, mode: str = "mrphs", truncate: bool = False, include_modifiees: bool = False
    ) -> str:
        """Convert a base phrase to a text.

        Args:
            bp: A base phrase.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            truncate: If true, adjunct words are truncated.
            include_modifiees: If true, parents are used to construct a compound phrase.
        """
        assert mode in {"mrphs", "reps"}
        if bp.omitted_case:
            if bp.exophora:
                base = bp.exophora
            else:
                mrphs = self._truncate_mrphs(list(bp.tag.mrph_list()))
                base = self._format_mrphs(mrphs, mode, normalize=True)
            case = convert_katakana_to_hiragana(self.case)
            case = case if mode == "mrphs" else f"{case}/{case}"
            return f"[{base}]" if truncate else f"[{base} {case}]"
        else:
            mrphs = list(bp.tag.mrph_list())
            if include_modifiees:
                for parent_base_phrase in bp.modifiees():
                    mrphs += parent_base_phrase.tag.mrph_list()
            if truncate:
                mrphs = self._truncate_mrphs(mrphs)
                return self._format_mrphs(mrphs, mode, normalize=True)
            else:
                return self._format_mrphs(mrphs, mode, normalize=False)

    @staticmethod
    def _truncate_mrphs(mrphs: List[Morpheme]) -> List[Morpheme]:
        """Truncate a list of morphemes.

        Args:
            mrphs: A list of morphemes.
        """
        content_mrphs = []
        seen_content_word = False
        for mrph in mrphs:
            is_content_word = mrph.hinsi not in {"助詞", "特殊", "判定詞"}
            if not is_content_word and seen_content_word:
                break
            seen_content_word = seen_content_word or is_content_word
            content_mrphs.append(mrph)
        return content_mrphs

    @staticmethod
    def _format_mrphs(mrphs: List[Morpheme], mode: str, normalize: bool = False) -> str:
        """Convert a list of morphemes to a text.

        Args:
            mrphs: A list of morphemes.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            normalize: If true, the last content word will be normalized.
        """
        assert mode in {"mrphs", "reps"}
        if mode == "reps":
            return " ".join(mrph.repname or f"{mrph.midasi}/{mrph.midasi}" for mrph in mrphs)
        else:
            if normalize:
                # Change the last morpheme to its infinitive (i.e., genkei).
                # Strip the return string for the case that len(mrphs) == 1.
                return (" ".join(mrph.midasi for mrph in mrphs[:-1]) + " " + mrphs[-1].genkei).strip()
            else:
                return " ".join(mrph.midasi for mrph in mrphs)

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            surf=self.surf,
            normalized_surf=self.normalized_surf,
            mrphs=self.mrphs,
            normalized_mrphs=self.normalized_mrphs,
            reps=self.reps,
            normalized_reps=self.normalized_reps,
            head_reps=self.head_reps,
            eid=self.eid,
            flag=self.flag,
            sdist=self.sdist,
            adnominal_event_ids=self.adnominal_event_ids,
            sentential_complement_event_ids=self.sentential_complement_event_ids,
            children=self.children,
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Argument, case: {self.case}, surf: {self.surf}>"


class ArgumentBuilder(Builder):
    def __call__(self, pas: "PAS", case: str, arg: PyknpArgument) -> Argument:
        argument = Argument(pas, case, arg.eid, arg.flag, arg.sdist, arg)
        pas.arguments[case].append(argument)
        return argument


class JsonArgumentBuilder(Builder):
    def __call__(self, pas: "PAS", case: str, dump: dict) -> Argument:
        argument = Argument(pas, case, dump["eid"], dump["flag"], dump["sdist"])
        argument._surf = dump["surf"]
        argument._normalized_surf = dump["normalized_surf"]
        argument._mrphs = dump["mrphs"]
        argument._normalized_mrphs = dump["normalized_mrphs"]
        argument._reps = dump["reps"]
        argument._normalized_reps = dump["normalized_reps"]
        argument._head_reps = dump["head_reps"]
        argument._children = dump["children"]
        argument._adnominal_event_ids = dump["adnominal_event_ids"]
        argument._sentential_complement_event_ids = dump["sentential_complement_event_ids"]
        pas.arguments[case].append(argument)
        return argument


class ArgumentsBuilder(Builder):
    def __call__(self, pas: "PAS") -> Dict[str, List[Argument]]:
        arguments: Dict[str, List[Argument]] = collections.defaultdict(list)
        if pas.pas:
            for case, args in sorted(pas.pas.arguments.items(), key=lambda x: PAS_ORDER.get(x[0], 99)):
                for arg in sorted(args, key=lambda _arg: (pas.ssid - _arg.sdist, _arg.tid)):
                    arguments[case].append(ArgumentBuilder()(pas, case, arg))
        return arguments


class JsonArgumentsBuilder(Builder):
    def __call__(self, pas: "PAS", dump: dict) -> Dict[str, List[Argument]]:
        arguments: Dict[str, List[Argument]] = collections.defaultdict(list)
        for case, arguments_dump in dump.items():
            for argument_dump in arguments_dump:
                arguments[case].append(JsonArgumentBuilder()(pas, case, argument_dump))
        return arguments



import re
from logging import getLogger
from typing import TYPE_CHECKING, List, Optional

from pyknp import Tag

from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event

logger = getLogger(__name__)


class Relation(Component):
    """A relation connects two events.
    Relations fall into two major divisions: syntactic and discourse relations. Syntactic relations can be used by
    application developers to, for example, construct a larger information unit by merging a modifier event to the
    modifiee, while discourse relations offer more pragmatic information, paving the way for deep language
    understanding.

    Attributes:
        modifier (Event): A modifier event.
        head (Event): A head event.
        label (str): A relation label. Syntactic relation labels include "連体修飾 (adnominal relation,"
            "補文 (sentential complement," "並列 (parallel)", and "係り受け (dependency)." On the other hand,
            discourse relation labels include "原因・理由 (cause/reason," "目的 (purpose," "条件 (condition,"
            "根拠 (ground," "対比 (contrast," and "逆接 (concession)."
        surf (str): A surface string.
        head_tid (int): A tag ID.
        reliable (bool): If true, a syntactic dependency is not ambiguous.
    """

    def __init__(self, modifier: "Event", head: "Event", label: str, surf: str, head_tid: int, reliable: bool):
        self.modifier: Optional[Event] = modifier
        self.head: Optional[Event] = head
        self.label: str = label
        self.surf: str = surf
        self.head_tid: int = head_tid
        self.reliable: bool = reliable

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            event_id=self.head.evid, label=self.label, surf=self.surf, reliable=self.reliable, head_tid=self.head_tid
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Relation, label: {self.label}, modifier_evid: {self.modifier.evid}, head_evid: {self.head.evid}>"


def filter_relations(
    relations: List[Relation], labels: List[str] = None, head_tids: List[int] = None
) -> List[Relation]:
    """Filter relations.

    Args:
        relations: A list of relations.
        labels: A list of valid labels.
        head_tids: A list of valid head tag IDs.
    """
    ret = []
    for relation in relations:
        if isinstance(labels, list) and relation.label not in labels:
            continue
        if isinstance(head_tids, list) and relation.head_tid not in head_tids:
            continue
        ret.append(relation)
    return ret


class RelationBuilder:
    def __call__(
        self, modifier: "Event", head: "Event", label: str, surf: str = "", head_tid: int = -1, reliable: bool = False
    ) -> Relation:
        relation = Relation(modifier, head, label, surf, head_tid, reliable)
        modifier.outgoing_relations.append(relation)
        head.incoming_relations.append(relation)
        return relation


class JsonRelationBuilder(Builder):
    def __call__(self, modifier_evid: int, head_evid: int, dump: dict) -> Relation:
        modifier = Builder.evid_event_map[modifier_evid]
        head = Builder.evid_event_map[head_evid]
        relation = Relation(modifier, head, dump["label"], dump["surf"], dump["head_tid"], dump["reliable"])
        modifier.outgoing_relations.append(relation)
        head.incoming_relations.append(relation)
        return relation


class RelationsBuilder(Builder):
    def __call__(self, event: "Event") -> List[Relation]:
        relations: List[Relation] = []
        for relation in self._get_outgoing_relations(event):
            relations.append(relation)
        return relations

    def _get_outgoing_relations(self, event: "Event") -> List[Relation]:
        relations: List[Relation] = []

        parent_event = self._find_parent(event)
        if parent_event:
            event.parent = parent_event

        # Dependency ambiguity.
        if event.parent:
            reliable = [event.evid, event.parent.evid] == [event_.evid for event_ in event.sentence.events][-2:]
        else:
            reliable = False

        # Adnominal.
        if event.parent and event.end.features["節-区切"] == "連体修飾":
            relations.append(
                RelationBuilder()(event, event.parent, "連体修飾", head_tid=event.end.parent_id, reliable=reliable)
            )

        # Sentential complement.
        if event.parent and event.end.features["節-区切"] == "補文":
            relations.append(
                RelationBuilder()(event, event.parent, "補文", head_tid=event.end.parent_id, reliable=reliable)
            )

        # Discourse relation.
        if not relations:
            for discourse_relation in re.findall("<談話関係[;:](.+?)>", event.end.fstring):
                tmp, label = discourse_relation.split(":")
                sdist, tid, sid = tmp.split("/")
                head_event = Builder.stid_event_map.get((event.ssid + int(sdist), int(tid)), None)
                if head_event:
                    relations.append(RelationBuilder()(event, head_event, f"談話関係:{label}"))

        # Clausal function.
        if not relations and event.parent:
            for clause_function in re.findall("<節-機能-(.+?)>", event.end.fstring):
                if ":" in clause_function:
                    label, surf = clause_function.split(":")
                else:
                    label, surf = clause_function, ""
                relations.append(
                    RelationBuilder()(
                        event, event.parent, label, surf=surf, head_tid=event.end.parent_id, reliable=reliable
                    )
                )

        # Clausal parallel relation.
        if not relations and event.parent:
            if event.end.dpndtype == "P":
                relations.append(RelationBuilder()(event, event.parent, "並列", reliable=reliable))

        # Clausal dependency.
        if not relations and event.parent:
            relations.append(RelationBuilder()(event, event.parent, "係り受け", reliable=reliable))
        return relations

    @staticmethod
    def _find_parent(event: "Event") -> Optional["Event"]:
        parent_tag: Optional[Tag] = event.head.parent
        while parent_tag:
            for parent_event_cand in filter(lambda event_: event.evid < event_.evid, event.sentence.events):
                if parent_tag.tag_id in {parent_event_cand.head.tag_id, parent_event_cand.end.tag_id}:
                    return parent_event_cand
            parent_tag = parent_tag.parent
        return None



from logging import getLogger
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from pyknp import Morpheme, Tag

from pyknp_eventgraph.base_phrase import BasePhrase, group_base_phrases
from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.features import Features, FeaturesBuilder, JsonFeaturesBuilder
from pyknp_eventgraph.helper import PAS_ORDER, convert_katakana_to_hiragana, convert_mrphs_to_surf
from pyknp_eventgraph.pas import PAS, JsonPASBuilder, PASBuilder
from pyknp_eventgraph.relation import Relation

if TYPE_CHECKING:
    from pyknp_eventgraph.sentence import Sentence

Morpheme_ = Union[str, Morpheme]

logger = getLogger(__name__)


class Event(Component):
    """Event is the basic information unit of EventGraph. Event is closely related to PAS but more
    application-oriented with respect to the following points:
      * Semantic heaviness: Some predicates are too semantically light for applications to treat as information units.
        EventGraph constrains an event to have a semantically heavy predicate.
      * Rich linguistic features: Linguistic features such as tense and modality are assigned to events.

    Attributes:
        sentence (:class:`.Sentence`): A sentence to which this event belongs.
        evid (int): A serial event ID.
        sid (str): An original sentence ID.
        ssid (int): A serial sentence ID.
        start (:class:`pyknp.knp.tag.Tag`, optional): A start tag.
        head (:class:`pyknp.knp.tag.Tag`, optional): A head tag.
        end (:class:`pyknp.knp.tag.Tag`, optional): An end tag.
        pas (PAS, optional): A predicate argument structure.
        outgoing_relations (List[Relation]): A list of relations where this event is the modifier.
        incoming_relations (List[Relation]): A list of relations where this event is the head.
        features (Features, optional): Linguistic features.
        parent (Event, optional): A parent event.
        children (List[Event]): A list of child events.
        head_base_phrase (BasePhrase, optional): A head basic phrase.
    """

    def __init__(
        self,
        sentence: "Sentence",
        evid: int,
        sid: str,
        ssid: int,
        start: Optional[Tag] = None,
        head: Optional[Tag] = None,
        end: Optional[Tag] = None,
    ):
        self.sentence: Sentence = sentence
        self.evid: int = evid
        self.sid: str = sid
        self.ssid: int = ssid
        self.start: Tag = start
        self.head: Tag = head
        self.end: Tag = end
        self.pas: Optional[PAS] = None
        self.outgoing_relations: List[Relation] = []
        self.incoming_relations: List[Relation] = []
        self.features: Optional[Features] = None
        self.parent: Optional[Event] = None
        self.children: List[Event] = []
        self.head_base_phrase: Optional[BasePhrase] = None

        self._surf = None
        self._surf_with_mark = None
        self._mrphs = None
        self._mrphs_with_mark = None
        self._normalized_mrphs = None
        self._normalized_mrphs_with_mark = None
        self._normalized_mrphs_without_exophora = None
        self._normalized_mrphs_with_mark_without_exophora = None
        self._reps = None
        self._reps_with_mark = None
        self._normalized_reps = None
        self._normalized_reps_with_mark = None
        self._content_rep_list = None

    @property
    def event_id(self) -> int:
        """An alias to evid."""
        return self.evid

    @property
    def surf(self) -> str:
        """A surface string."""
        if self._surf is None:
            self._surf = self.surf_()
        return self._surf

    @property
    def surf_with_mark(self) -> str:
        """A surface string with marks."""
        if self._surf_with_mark is None:
            self._surf_with_mark = self.surf_with_mark_()
        return self._surf_with_mark

    @property
    def mrphs(self) -> str:
        """A tokenized surface string."""
        if self._mrphs is None:
            self._mrphs = self.mrphs_()
        return self._mrphs

    @property
    def mrphs_with_mark(self) -> str:
        """A tokenized surface string with marks."""
        if self._mrphs_with_mark is None:
            self._mrphs_with_mark = self.mrphs_with_mark_()
        return self._mrphs_with_mark

    @property
    def normalized_mrphs(self) -> str:
        """A tokenized/normalized surface string."""
        if self._normalized_mrphs is None:
            self._normalized_mrphs = self.normalized_mrphs_()
        return self._normalized_mrphs

    @property
    def normalized_mrphs_with_mark(self) -> str:
        """A tokenized/normalized surface string with marks."""
        if self._normalized_mrphs_with_mark is None:
            self._normalized_mrphs_with_mark = self.normalized_mrphs_with_mark_()
        return self._normalized_mrphs_with_mark

    @property
    def normalized_mrphs_without_exophora(self) -> str:
        """A tokenized/normalized surface string without exophora."""
        if self._normalized_mrphs_without_exophora is None:
            self._normalized_mrphs_without_exophora = self.normalized_mrphs_without_exophora_()
        return self._normalized_mrphs_without_exophora

    @property
    def normalized_mrphs_with_mark_without_exophora(self) -> str:
        """A tokenized/normalized surface string with marks but without exophora."""
        if self._normalized_mrphs_with_mark_without_exophora is None:
            self._normalized_mrphs_with_mark_without_exophora = self.normalized_mrphs_with_mark_without_exophora_()
        return self._normalized_mrphs_with_mark_without_exophora

    @property
    def reps(self) -> str:
        """A representative string."""
        if self._reps is None:
            self._reps = self.reps_()
        return self._reps

    @property
    def reps_with_mark(self) -> str:
        """A representative string with marks."""
        if self._reps_with_mark is None:
            self._reps_with_mark = self.reps_with_mark_()
        return self._reps_with_mark

    @property
    def normalized_reps(self) -> str:
        """A normalized representative string."""
        if self._normalized_reps is None:
            self._normalized_reps = self.normalized_reps_()
        return self._normalized_reps

    @property
    def normalized_reps_with_mark(self) -> str:
        """A normalized representative string with marks."""
        if self._normalized_reps_with_mark is None:
            self._normalized_reps_with_mark = self.normalized_reps_with_mark_()
        return self._normalized_reps_with_mark

    @property
    def content_rep_list(self) -> List[str]:
        """A list of content words."""
        if self._content_rep_list is None:
            self._content_rep_list = self.content_rep_list_()
        return self._content_rep_list

    def surf_(self, include_modifiers: bool = False) -> str:
        """A surface string.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return convert_mrphs_to_surf(self.mrphs_(include_modifiers))

    def surf_with_mark_(self, include_modifiers: bool = False) -> str:
        """A surface string with marks.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return convert_mrphs_to_surf(self.mrphs_with_mark_(include_modifiers))

    def mrphs_(self, include_modifiers: bool = False) -> str:
        """A tokenized surface string.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=False, add_mark=False, include_modifiers=include_modifiers)

    def mrphs_with_mark_(self, include_modifiers: bool = False) -> str:
        """A tokenized surface string with marks.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=False, add_mark=True, include_modifiers=include_modifiers)

    def normalized_mrphs_(self, include_modifiers: bool = False) -> str:
        """A tokenized/normalized surface string.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=True, add_mark=False, include_modifiers=include_modifiers)

    def normalized_mrphs_with_mark_(self, include_modifiers: bool = False) -> str:
        """A tokenized/normalized surface string with marks.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=True, add_mark=True, include_modifiers=include_modifiers)

    def normalized_mrphs_without_exophora_(self, include_modifiers: bool = False) -> str:
        """A tokenized/normalized surface string without exophora.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=True, add_mark=False, exclude_exophora=True, include_modifiers=include_modifiers)

    def normalized_mrphs_with_mark_without_exophora_(self, include_modifiers: bool = False) -> str:
        """A tokenized/normalized surface string with marks but without exophora.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text(truncate=True, add_mark=True, exclude_exophora=True, include_modifiers=include_modifiers)

    def reps_(self, include_modifiers: bool = False) -> str:
        """A representative string.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text("reps", truncate=False, add_mark=False, include_modifiers=include_modifiers)

    def reps_with_mark_(self, include_modifiers: bool = False) -> str:
        """A representative string with marks.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text("reps", truncate=False, add_mark=True, include_modifiers=include_modifiers)

    def normalized_reps_(self, include_modifiers: bool = False) -> str:
        """A normalized representative string.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text("reps", truncate=True, add_mark=False, include_modifiers=include_modifiers)

    def normalized_reps_with_mark_(self, include_modifiers: bool = False) -> str:
        """A normalized representative string with marks.

        Args:
            include_modifiers: If true, tokens of events that modify this event will be included.
        """
        return self._to_text("reps", truncate=True, add_mark=True, include_modifiers=include_modifiers)

    def content_rep_list_(self) -> List[str]:
        """A list of content words."""
        content_rep_list = []
        for bp in self._collect_base_phrases():
            if bp.tag is None:
                continue
            for mrph in bp.tag.mrph_list():
                if "<内容語>" in mrph.fstring or "<準内容語>" in mrph.fstring:
                    content_rep_list.append(mrph.repname or f"{mrph.midasi}/{mrph.midasi}")
        return content_rep_list

    def _to_text(
        self,
        mode: str = "mrphs",
        truncate: bool = False,
        add_mark: bool = False,
        exclude_exophora: bool = False,
        include_modifiers: bool = False,
        exclude_adnominal: bool = False,
    ) -> str:
        """Convert this event to a text.

        Args:
            mode: A type of token representation, which can take either "mrphs" or "reps".
            truncate: If true, adjunct words are truncated.
            add_mark: If true, special marks are added.
            exclude_exophora: If true, exophora will not be used.
            include_modifiers: If true, tokens of events that modify this event will be included.
            exclude_adnominal: If true, base phrases modified by this event will be excluded.
        """
        assert mode in {"mrphs", "reps"}

        # Create a list of base phrases to show.
        grouped_bps = group_base_phrases(
            self._collect_base_phrases(exclude_exophora=exclude_exophora, exclude_adnominal=exclude_adnominal)
        )

        # Create a list of morphemes.
        grouped_mrphs = [[morpheme for bp in bps for morpheme in bp.morphemes] for bps in grouped_bps]

        # Truncate the morphemes.
        truncated_pos = self._find_truncated_position(grouped_bps)
        if truncate:
            grouped_mrphs = grouped_mrphs[: truncated_pos[0] + 1]
            grouped_mrphs[-1] = grouped_mrphs[-1][: truncated_pos[1] + 1]

        # Create a map from a position to a string to be inserted.
        additional_texts = self._get_additional_texts(
            grouped_bps=grouped_bps,
            grouped_mrphs=grouped_mrphs,
            mode=mode,
            add_mark=add_mark,
            normalize=truncate,
            truncated_pos=truncated_pos,
            include_modifiers=include_modifiers,
            exclude_exophora=exclude_exophora,
        )

        return self._format_grouped_mrphs(
            grouped_mrphs=grouped_mrphs, mode=mode, normalize=truncate, additional_texts=additional_texts
        )

    def _collect_base_phrases(
        self,
        exclude_exophora: bool = False,
        exclude_adnominal: bool = False,
    ) -> List[BasePhrase]:
        """Collect base phrases belonging to this event.

        Args:
            exclude_exophora: If true, exophora will be excluded.
            exclude_adnominal: If true, base phrases modified by this event will be excluded.

        Returns:
            A list of base phrases that belong to this event.
        """
        # Collect head base phrases.
        head_bps = [self.pas.predicate.head_base_phrase]
        for args in self.pas.arguments.values():
            for arg in args:
                if arg.head_base_phrase.omitted_case:
                    if exclude_exophora and arg.head_base_phrase.exophora:
                        # e.g., [著者 が]
                        continue
                    if exclude_adnominal and arg.head_base_phrase.tag == self.pas.predicate.head_base_phrase.tag.parent:
                        # e.g., [車が] of "[車が] 高速道路を低速で走る" -> "車は危ない"
                        continue
                    head_bps.append(arg.head_base_phrase)
                    continue
                if arg.head_base_phrase.is_event_head or arg.head_base_phrase.is_event_end:
                    continue
                if arg.head_base_phrase.tag.tag_id > self.end.tag_id:
                    continue
                head_bps.append(arg.head_base_phrase)
        return sorted(list(set(bp for head_bp in head_bps for bp in head_bp.to_list())))

    def _find_truncated_position(self, grouped_bps: List[List[BasePhrase]]) -> Tuple[int, int]:
        """Find a position just before adjunct words start.

        Args:
            grouped_bps: A list of base phrases grouped by bunsetsu IDs.

        Returns:
            A position just before adjunct words start.
        """
        seen_head = False
        for group_index, bps in enumerate(grouped_bps):
            # Ignore base phrases of a omitted case because they never become a predicate.
            if any(bp.omitted_case for bp in bps):
                continue
            mrph_index_offset = 0
            for bp in bps:
                # Skip base phrases until the current base phrase reaches to the predicate's head base phrase.
                seen_head = seen_head or bp == self.pas.predicate.head_base_phrase
                if not seen_head:
                    mrph_index_offset += len(bp.morphemes)
                    continue
                # Find a position to be truncated.
                for mrph_index, mrph in reversed(list(enumerate(bp.morphemes))):
                    if (
                        mrph.hinsi == "助動詞"
                        and mrph.genkei == "です"
                        and 0 < mrph_index
                        and bp.morphemes[mrph_index - 1].hinsi == "形容詞"
                    ):
                        # adjective + 'です' -> ignore 'です' (e.g., 美しいです -> 美しい)
                        return group_index, mrph_index_offset + mrph_index - 1

                    if (
                        mrph.hinsi == "判定詞"
                        and mrph.midasi == "じゃ"
                        and 0 < mrph_index
                        and "<活用語>" in bp.morphemes[mrph_index - 1].fstring
                    ):
                        # adjective or verb +'じゃん' -> ignore 'じゃん' (e.g., 使えないじゃん -> 使えない)
                        return group_index, mrph_index_offset + mrph_index - 1

                    if ("<活用語>" in mrph.fstring or "<用言意味表記末尾>" in mrph.fstring) and mrph.genkei not in {"のだ", "んだ"}:
                        # Check the last word with conjugation except some meaningless words.
                        return group_index, mrph_index_offset + mrph_index
                mrph_index_offset += len(bp.morphemes)
        return len(grouped_bps) - 1, sum(len(bp.tag.mrph_list()) for bp in grouped_bps[-1]) - 1

    @staticmethod
    def _get_additional_texts(
        grouped_bps: List[List[BasePhrase]],
        grouped_mrphs: List[List[Morpheme_]],
        mode: str,
        add_mark: bool,
        normalize: bool,
        truncated_pos: Tuple[int, int],
        include_modifiers: bool,
        exclude_exophora: bool,
    ) -> Dict[Tuple[int, int, str], str]:
        """Get a mapping from a position to a mark.

        Args:
            grouped_bps: A list of base phrases grouped by bunsetsu IDs.
            grouped_mrphs: A list of morphemes grouped by bunsetsu IDs.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            add_mark: If true, add special marks.
            normalize: If true, the last content word will be normalized.
            truncated_pos: A position just before adjunct words start.
            include_modifiers: If true, tokens of events that modify this event will be included.
            exclude_exophora: If true, exophora will not be used.

        Returns:
            A mapping from positions to marks.
        """
        additional_texts: Dict[Tuple[int, int, str], str] = {}  # (group_index, mrph_index, "start" or "end") -> text

        def get_event_str(event: "Event") -> str:
            return (
                event._to_text(
                    mode,
                    truncate=False,
                    add_mark=add_mark,
                    exclude_exophora=exclude_exophora,
                    include_modifiers=include_modifiers,
                    exclude_adnominal=True,
                )
                .replace(" (", "")
                .replace(")", "")
            )

        last_tid = -1
        for group_index, (bps, mrphs) in enumerate(zip(grouped_bps, grouped_mrphs)):
            start_pos = (group_index, 0, "start")
            end_pos = (group_index, len(mrphs) - 1, "end")

            is_omitted = any(bp.omitted_case for bp in bps)
            if is_omitted:
                additional_texts[start_pos] = "["
                additional_texts[end_pos] = "]"
                continue

            if add_mark or include_modifiers:
                adnominal_events = sorted([e for bp in bps for e in bp.adnominal_events], key=lambda e: e.evid)
                if adnominal_events:
                    if include_modifiers:
                        additional_texts[start_pos] = " ".join(get_event_str(e) for e in adnominal_events)
                    else:
                        additional_texts[start_pos] = "▼"
                sentential_complement_events = sorted(
                    [e for bp in bps for e in bp.sentential_complement_events], key=lambda e: e.evid
                )
                if sentential_complement_events:
                    if include_modifiers:
                        additional_texts[start_pos] = " ".join(get_event_str(e) for e in sentential_complement_events)
                    else:
                        additional_texts[start_pos] = "■"

            if add_mark:
                mrph_index = 0
                for bp in bps:
                    pos = (group_index, mrph_index, "start")
                    if last_tid != -1 and last_tid + 1 != bp.tid and pos not in additional_texts:
                        additional_texts[pos] = "|"
                    last_tid = bp.tid
                    mrph_index += len(bp.tag.mrph_list())

        last_pos = (len(grouped_mrphs) - 1, len(grouped_mrphs[-1]) - 1)
        if add_mark and not normalize and truncated_pos != last_pos:
            additional_texts[(truncated_pos[0], truncated_pos[1], "end")] = "("
            additional_texts[(len(grouped_mrphs) - 1, len(grouped_mrphs[-1]) - 1, "end")] = ")"

        return additional_texts

    @staticmethod
    def _format_grouped_mrphs(
        grouped_mrphs: List[List[Morpheme_]],
        mode: str,
        normalize: bool,
        additional_texts: Dict[Tuple[int, int, str], str],
    ) -> str:
        """Format a list of morphemes grouped by bunsetsu IDs to create a text.

        Args:
            grouped_mrphs: A list of morphemes grouped by bunsetsu IDs.
            mode: A type of token representation, which can take either "mrphs" or "reps".
            normalize: If true, the last content word will be normalized.
            additional_texts: A mapping from positions to marks.
        """
        assert mode in {"mrphs", "reps"}

        ret = []
        for group_index, mrphs in enumerate(grouped_mrphs):
            for mrph_index, mrph in enumerate(mrphs):
                if (group_index, mrph_index, "start") in additional_texts:
                    ret.append(additional_texts[(group_index, mrph_index, "start")])

                if isinstance(mrph, str):
                    if mrph in PAS_ORDER:
                        case = convert_katakana_to_hiragana(mrph)
                        ret.append(case if mode == "mrphs" else f"{case}/{case}")
                    else:
                        ret.append(mrph)
                else:
                    if mode == "reps":
                        ret.append(mrph.repname or f"{mrph.midasi}/{mrph.midasi}")
                    else:
                        if normalize and (group_index, mrph_index) == (len(grouped_mrphs) - 1, len(mrphs) - 1):
                            if mrph.hinsi == "助動詞" and mrph.genkei == "ぬ":
                                # Exception: prevent transforming "できません" into "できませぬ".
                                ret.append(mrph.midasi)
                            else:
                                ret.append(mrph.genkei)
                        else:
                            ret.append(mrph.midasi)

                if (group_index, mrph_index, "end") in additional_texts:
                    ret.append(additional_texts[(group_index, mrph_index, "end")])

        return " ".join(ret).replace("[ ", "[").replace(" ]", "]").replace("( ", "(").replace(" )", ")")

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            event_id=self.evid,
            sid=self.sid,
            ssid=self.ssid,
            rel=[r.to_dict() for r in self.outgoing_relations],
            surf=self.surf,
            surf_with_mark=self.surf_with_mark,
            mrphs=self.mrphs,
            mrphs_with_mark=self.mrphs_with_mark,
            normalized_mrphs=self.normalized_mrphs,
            normalized_mrphs_with_mark=self.normalized_mrphs_with_mark,
            normalized_mrphs_without_exophora=self.normalized_mrphs_without_exophora,
            normalized_mrphs_with_mark_without_exophora=self.normalized_mrphs_with_mark_without_exophora,
            reps=self.reps,
            reps_with_mark=self.reps_with_mark,
            normalized_reps=self.normalized_reps,
            normalized_reps_with_mark=self.normalized_reps_with_mark,
            content_rep_list=self.content_rep_list,
            pas=self.pas.to_dict(),
            features=self.features.to_dict(),
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Event, evid: {self.evid}, surf: {self.surf}>"


class EventBuilder(Builder):
    def __call__(self, sentence: "Sentence", start: Tag, head: Tag, end: Tag):
        event = Event(sentence, Builder.evid, sentence.sid, sentence.ssid, start, head, end)
        PASBuilder()(event)
        FeaturesBuilder()(event)
        sentence.events.append(event)
        Builder.evid += 1
        for tid in range(start.tag_id, end.tag_id + 1):
            Builder.stid_event_map[(sentence.ssid, tid)] = event
        return event


class JsonEventBuilder(Builder):
    def __call__(self, sentence: "Sentence", dump: dict) -> Event:
        event = Event(sentence, Builder.evid, sentence.sid, sentence.ssid)
        event._surf = dump["surf"]
        event._surf_with_mark = dump["surf_with_mark"]
        event._mrphs = dump["mrphs"]
        event._mrphs_with_mark = dump["mrphs_with_mark"]
        event._normalized_mrphs = dump["normalized_mrphs"]
        event._normalized_mrphs_with_mark = dump["normalized_mrphs_with_mark"]
        event._normalized_mrphs_without_exophora = dump["normalized_mrphs_without_exophora"]
        event._normalized_mrphs_with_mark_without_exophora = dump["normalized_mrphs_with_mark_without_exophora"]
        event._reps = dump["reps"]
        event._reps_with_mark = dump["reps_with_mark"]
        event._normalized_reps = dump["normalized_reps"]
        event._normalized_reps_with_mark = dump["normalized_reps_with_mark"]
        event._content_rep_list = dump["content_rep_list"]
        JsonPASBuilder()(event, dump["pas"])
        JsonFeaturesBuilder()(event, dump["features"])
        sentence.events.append(event)
        Builder.evid += 1
        Builder.evid_event_map[event.evid] = event
        return event



"""The base class of EventGraph components."""
from typing import TYPE_CHECKING, ClassVar, Dict, NoReturn, Tuple

from pyknp import Tag

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event


class Builder:
    """The base of builders."""

    ssid: int = 0
    evid: int = 0

    evid_event_map: ClassVar[Dict[int, "Event"]] = {}
    stid_event_map: ClassVar[Dict[Tuple[int, int], "Event"]] = {}
    stid_bid_map: ClassVar[Dict[Tuple[int, int], int]] = {}
    stid_tag_map: ClassVar[Dict[Tuple[int, int], Tag]] = {}

    @classmethod
    def reset(cls) -> NoReturn:
        cls.ssid = 0
        cls.evid = 0
        cls.stid_event_map = {}
        cls.stid_bid_map = {}
        cls.stid_tag_map = {}



import collections
import itertools
import os
from logging import getLogger
from typing import List

import graphviz

from pyknp_eventgraph import EventGraph
from pyknp_eventgraph.eventgraph import Event, Relation
from pyknp_eventgraph.helper import PAS_ORDER

logger = getLogger(__name__)


def make_image(evg: EventGraph, output: str, with_detail: bool = True, with_original_text: bool = True):
    """Visualize an EventGraph.

    Args:
        evg (EventGraph): An EventGraph.
        output (str): Path to an output file. The file extension must be '.svg'.
        with_detail (bool): If true, detail information will be included.
        with_original_text (bool): If true, original sentences will be included.
    """
    output, ext = os.path.splitext(output)
    assert ext == ".svg", 'the extension of the output file must be ".svg"'

    # Group sentences and events by their sentence IDs
    sentences = {k: list(v) for k, v in itertools.groupby(evg.sentences, key=lambda x: x.sid.rsplit("-", 1)[0])}
    events = {k: list(v) for k, v in itertools.groupby(evg.events, key=lambda x: x.sid.rsplit("-", 1)[0])}

    # Create a base image
    g = graphviz.Digraph("G", format="svg")
    g.attr("graph", ranksep="0", margin="0", pad="0")

    num_cluster = 0
    for did in sentences.keys():
        doc_sentences = sentences.get(did, [])
        doc_events = events.get(did, [])

        if with_original_text:
            with g.subgraph(name=f"cluster_{num_cluster}") as h:
                h.attr("graph", style="invis")
                h.node(
                    name=f"head_{num_cluster}",
                    label="\\l".join(sentence.surf for sentence in doc_sentences) + "\\l",
                    shape="plaintext",
                )
                h.node(name=f"cluster_{num_cluster}_top", label="", shape="none", width="0")
            num_cluster += 1

        sent_events_list = _split_events_by_sid(doc_events)  # too long sentences are split
        for row, sent_events in enumerate(sent_events_list):
            with g.subgraph(name=f"cluster_{num_cluster}") as c:
                c.attr("graph", style="invis")
                for event in reversed(sent_events):
                    node = Node(event)
                    c.node(name=node.name, label=node.to_string(with_detail), shape="box", labelloc="b", height="0")
                c.node(name=f"cluster_{num_cluster}_top", label="", shape="none", width="0")
            num_cluster += 1

        for event in doc_events:
            for relation in event.outgoing_relations:
                edge = Edge(relation)
                g.edge(
                    tail_name=edge.modifier_node_name,
                    head_name=edge.head_node_name,
                    label=edge.to_string(),
                    weight="1",
                    constraint="false",
                )

    # align clusters vertically
    for i in range(num_cluster - 1):
        g.edge(tail_name=f"cluster_{i}_top", head_name=f"cluster_{i + 1}_top", style="invis")

    output_dir = os.path.abspath(os.path.dirname(output))
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    logger.debug("Render an image")
    g.render(output, cleanup=True)

    logger.debug("Successfully constructed visualization")


def _split_events_by_sid(events: List[Event], max_length: int = 4) -> List[List[Event]]:
    """Group events by their sentence IDs.

    Args:
        events: A list of events.
        max_length: A maximum number of events which are written in the same row.

    Returns:
        A list of lists of events.
    """
    ssid_events_map = collections.defaultdict(list)
    for event in events:
        ssid_events_map[event.ssid].append(event)
    split_events = []
    for ssid, sent_events in sorted(ssid_events_map.items(), key=lambda x: x[0]):
        for i in range(0, len(sent_events), max_length):
            split_events.append(sent_events[i : i + max_length])
    return split_events


class Node:
    def __init__(self, event: Event):
        self.event = event

    @property
    def name(self) -> str:
        """The name of this node."""
        return f"event_{self.event.evid}"

    @property
    def surf(self) -> str:
        """The surface string of this node."""
        return self.event.surf_with_mark

    @property
    def pas(self) -> str:
        """The PAS of this node."""
        pred = self.event.pas.predicate.standard_reps
        if self.event.pas.predicate.type_:
            pred += f":{self.event.pas.predicate.type_}"
        args = []
        for case in sorted(self.event.pas.arguments, key=lambda x: PAS_ORDER.get(x, 99)):
            arg = self.event.pas.arguments[case][0]
            if "外の関係" not in case:
                args.append(f"{arg.head_reps}:{case}")
        return ", ".join([pred] + args)

    @property
    def features(self) -> str:
        """The features of this node."""
        features = []
        if self.event.features.negation:
            features.append("否定")
        if self.event.features.tense:
            features.append(f"時制:{self.event.features.tense}")
        for modality in self.event.features.modality:
            features.append(f"モダリティ:{modality}")
        return ", ".join(features)

    def to_string(self, with_detail: bool) -> str:
        """Return the string.

        Args:
            with_detail: Whether to include the detail information.

        Returns:
            The string of a given event.
        """
        content = ""
        if with_detail:
            surf = self.surf
            if surf.endswith(")"):
                main, adjunct = surf[:-1].rsplit("(", 1)
                surf = f'{main.strip()}<font color="gray">{adjunct.strip()}</font>'
            content += f'<tr><td align="left">[surf] {surf}</td></tr>'
            pas = self.pas
            content += f'<tr><td align="left">[pas] {pas}</td></tr>'
            features = self.features
            if features:
                content += f'<tr><td align="left">[features] {features}</td></tr>'
        else:
            surf = self.surf
            if self.surf.endswith(")"):
                main, adjunct = self.surf[:-1].rsplit("(", 1)
                surf = f'{main.strip()}<font color="gray">{adjunct.strip()}</font>'
            content += f'<tr><td align="left">{surf}</td></tr>'
        return f'<<table border="0" cellborder="0" cellspacing="1">{content}</table>>'


class Edge:
    def __init__(self, relation: Relation):
        self.relation = relation

    @property
    def modifier_node_name(self) -> str:
        """The name of the modifier node."""
        return Node(self.relation.modifier).name

    @property
    def head_node_name(self) -> str:
        """The name of the head node."""
        return Node(self.relation.head).name

    def to_string(self) -> str:
        """Return the string."""
        label = self.relation.label.replace("談話関係", "談").replace("連体修飾", "▼").replace("補文", "■").replace("係り受け", "")
        out = label
        if out and self.relation.surf:
            out += f":{self.relation.surf}"
        return f"   {out}    "



import collections
from logging import getLogger
from typing import TYPE_CHECKING, Dict, List, Optional

from pyknp.knp.pas import Pas as PyknpPAS

from pyknp_eventgraph.argument import Argument, ArgumentsBuilder, JsonArgumentsBuilder
from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.predicate import JsonPredicateBuilder, Predicate, PredicateBuilder

if TYPE_CHECKING:
    from pyknp_eventgraph.event import Event

logger = getLogger(__name__)


class PAS(Component):
    """A PAS is the core of an event.

    Attributes:
        event (Event): An event that this PAS belongs.
        sid (str): An original sentence ID.
        ssid (int): A serial sentence ID.
        pas (:class:`pyknp.knp.pas.Pas`, optional): A PAS object in pyknp.
        predicate (Predicate): A predicate.
        arguments (Dict[str, List[Argument]]): A mapping of a case to arguments.
    """

    def __init__(self, event: "Event", pas: Optional[PyknpPAS] = None):
        self.event: Event = event
        self.sid: str = event.sid
        self.ssid: int = event.ssid
        self.pas: Optional[PyknpPAS] = pas
        self.predicate: Optional[Predicate] = None
        self.arguments: Optional[Dict[str, List[Argument]]] = collections.defaultdict(list)

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(
            predicate=self.predicate.to_dict(),
            argument={
                case: [argument.to_dict() for argument in argument_list if argument.to_dict()]
                for case, argument_list in self.arguments.items()
                if any(argument.to_dict() for argument in argument_list)
            },
        )

    def to_string(self) -> str:
        """Convert this object into a string."""
        predicate = self.predicate.standard_reps
        arguments = []
        for case, argument_list in self.arguments.items():
            for argument in argument_list:
                if argument.head_reps:
                    arguments.append(f"{case}: {argument.head_reps}")
        return f'<PAS, predicate: {predicate}, arguments: {"{" + ", ".join(arguments) + "}" if arguments else "None"}>'


class PASBuilder(Builder):
    def __call__(self, event: "Event") -> PAS:
        pas = PAS(event, event.head.pas)
        PredicateBuilder()(pas)
        ArgumentsBuilder()(pas)
        event.pas = pas
        return pas


class JsonPASBuilder(Builder):
    def __call__(self, event: "Event", dump: dict) -> PAS:
        pas = PAS(event)
        JsonPredicateBuilder()(pas, dump["predicate"])
        JsonArgumentsBuilder()(pas, dump["argument"])
        event.pas = pas
        return pas



import collections
from typing import TYPE_CHECKING, List, NoReturn, Optional, Tuple, Union

from pyknp import Morpheme, Tag

from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.helper import PAS_ORDER, convert_katakana_to_hiragana, get_parallel_tags
from pyknp_eventgraph.relation import filter_relations

if TYPE_CHECKING:
    from pyknp_eventgraph.argument import Argument
    from pyknp_eventgraph.event import Event
    from pyknp_eventgraph.predicate import Predicate


class BasePhrase(Component):
    """A wrapper of :class:`pyknp.knp.tag.Tag`, which allow exophora to be a base phrase.
    BasePhrase is a bidirectional linked list; each of base phrases has its parent and children.

    Attributes:
        event (Event): An event that has this base phrase.
        tag (Tag, optional): A tag.
        ssid (int): A serial sentence ID.
        bid (int): A serial bunsetsu ID.
        tid (int): A serial tag ID.
        is_child (bool): If true, this base phrase is a child of a head base phrase.
        exophora (str): An exophora.
        omitted_case (str): A omitted case.
        parent (BasePhrase, optional): A parent base phrase.
        children (List[BasePhrase]): A list of child base phrases.
    """

    def __init__(
        self,
        event: "Event",
        tag: Optional[Tag],
        ssid: int,
        bid: int,
        tid: int,
        is_child: bool = False,
        exophora: str = "",
        omitted_case: str = "",
    ):
        self.event = event
        self.tag: Optional[Tag] = tag
        self.ssid = ssid
        self.bid = bid
        self.tid = tid
        self.is_child = is_child
        self.exophora = exophora
        self.omitted_case = omitted_case
        self.parent: Optional["BasePhrase"] = None
        self.children: List["BasePhrase"] = []

        self._surf = None

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other: "BasePhrase"):
        assert isinstance(other, BasePhrase)
        return self.key == other.key

    def __lt__(self, other: "BasePhrase"):
        assert isinstance(other, BasePhrase)
        return self.key < other.key

    @property
    def morphemes(self) -> List[Union[str, Morpheme]]:
        mrphs = []
        if self.omitted_case:
            if self.exophora:
                mrphs.append(self.exophora)
            else:
                exists_content_word = False
                for mrph in self.tag.mrph_list():
                    is_content_word = mrph.hinsi not in {"助詞", "特殊", "判定詞"}
                    if not is_content_word and exists_content_word:
                        break
                    exists_content_word = exists_content_word or is_content_word
                    mrphs.append(mrph)
            mrphs.append(self.omitted_case)
        else:
            mrphs.extend(list(self.tag.mrph_list()))
        return mrphs

    @property
    def surf(self) -> str:
        """A surface string."""
        if self._surf is None:
            morphemes = self.morphemes
            if self.omitted_case:
                bases, case = morphemes[:-1], morphemes[-1]
                base = "".join(base if isinstance(base, str) else base.midasi for base in bases)
                case = convert_katakana_to_hiragana(case)
                self._surf = f"[{base}{case}]"
            else:
                self._surf = "".join(mrph.midasi for mrph in morphemes)
        return self._surf

    @property
    def key(self) -> Tuple[int, int, int, int]:
        """A key used for sorting."""
        return PAS_ORDER.get(self.omitted_case, 99), self.ssid, self.bid, self.tid

    @property
    def is_event_head(self) -> bool:
        """True if this base phrase is the head of an event."""
        return bool(self.tag and any("節-主辞" in tag.features for tag in [self.tag] + get_parallel_tags(self.tag)))

    @property
    def is_event_end(self) -> bool:
        """True if this base phrase is the end of an event."""
        return bool(self.tag and any("節-区切" in tag.features for tag in [self.tag] + get_parallel_tags(self.tag)))

    @property
    def adnominal_events(self) -> List["Event"]:
        """A list of events modifying this predicate (adnominal)."""
        if self.omitted_case:
            return []
        else:
            return [r.modifier for r in filter_relations(self.event.incoming_relations, ["連体修飾"], [self.tid])]

    @property
    def sentential_complement_events(self) -> List["Event"]:
        """A list of events modifying this predicate (sentential complement)."""
        if self.omitted_case:
            return []
        else:
            return [r.modifier for r in filter_relations(self.event.incoming_relations, ["補文"], [self.tid])]

    @property
    def root(self) -> "BasePhrase":
        """Return the root of this base phrase."""
        root_bp = self
        while root_bp.parent:
            root_bp = root_bp.parent
        return root_bp

    def to_list(self) -> List["BasePhrase"]:
        """Expand to a list."""
        return sorted(self.root.modifiers(include_self=True))

    def modifiees(self, include_self: bool = False) -> List["BasePhrase"]:
        """Return a list of base phrases modified by this base phrase.

        Args:
            include_self: If true, include this base phrase to the return.
        """
        modifiee_bps = [self] if include_self else []

        def add_modifiee(bp: BasePhrase):
            if bp.parent:
                modifiee_bps.append(bp.parent)
                add_modifiee(bp.parent)

        add_modifiee(self)
        return modifiee_bps

    def modifiers(self, include_self: bool = False) -> List["BasePhrase"]:
        """Return a list of base phrases modifying this base phrase.

        Args:
            include_self: If true, include this base phrase to the return.
        """
        modifier_bps = [self] if include_self else []

        def add_modifier(bp: BasePhrase):
            for child_bp in bp.children:
                modifier_bps.append(child_bp)
                add_modifier(child_bp)

        add_modifier(self)
        return sorted(modifier_bps)

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(ssid=self.ssid, bid=self.bid, tid=self.tid, surf=self.surf)

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<BasePhrase, ssid: {self.ssid}, bid: {self.bid}, tid: {self.tid}, surf: {self.surf}>"


def group_base_phrases(bps: List[BasePhrase]) -> List[List[BasePhrase]]:
    """Group base phrases by their bunsetsu IDs (bid).

    Args:
        bps: A list of base phrases.

    Returns:
        A list of base phrases grouped by bunsetsu IDs.
    """
    bucket = collections.defaultdict(list)
    for bp in sorted(bps):
        bucket[bp.key[:-1]].append(bp)  # bp.key[-1] is the tag id.
    return list(bucket.values())  # In Python 3.6+, dictionaries are insertion ordered.


class BasePhraseBuilder(Builder):
    def __call__(self, event: "Event"):
        # Greedily dispatch base phrases to arguments.
        argument_head_bps: List[BasePhrase] = []
        for args in event.pas.arguments.values():
            for arg in args:
                head = self.dispatch_head_base_phrase_to_argument(arg)
                argument_head_bps.append(head)
                if head.parent:
                    argument_head_bps.append(head.parent)

        # Resolve duplication.
        self._resolve_duplication(argument_head_bps)

        # Dispatch base phrases to a predicate.
        self.dispatch_head_base_phrase_to_predicate(event.pas.predicate, sentinels=argument_head_bps)

    def dispatch_head_base_phrase_to_argument(self, argument: "Argument") -> BasePhrase:
        event = argument.pas.event
        ssid = argument.pas.ssid - argument.arg.sdist
        tid = argument.arg.tid
        bid = Builder.stid_bid_map.get((ssid, tid), -1)
        tag = Builder.stid_tag_map.get((ssid, tid), None)

        if argument.arg.flag == "E":  # exophora
            head_bp = BasePhrase(event, None, ssid, bid, tid, exophora=argument.arg.midasi, omitted_case=argument.case)
        elif argument.arg.flag == "O":  # zero anaphora
            head_bp = BasePhrase(event, tag, ssid, bid, tid, omitted_case=argument.case)
        else:
            head_bp = BasePhrase(event, tag, ssid, bid, tid)
            self.add_children(head_bp, ssid)
            self.add_compound_phrase_component(head_bp, ssid)

        argument.head_base_phrase = head_bp
        return head_bp

    def dispatch_head_base_phrase_to_predicate(self, predicate: "Predicate", sentinels: List[BasePhrase]) -> BasePhrase:
        event = predicate.pas.event
        ssid = predicate.pas.event.ssid
        tid = predicate.head.tag_id
        bid = Builder.stid_bid_map.get((ssid, tid), -1)
        tag = Builder.stid_tag_map.get((ssid, tid), None)

        head_bp = BasePhrase(event, tag, ssid, bid, tid)
        self.add_children(head_bp, ssid, sentinels=sentinels)
        if predicate.pas.event.head != predicate.pas.event.end:
            next_tid = predicate.pas.event.end.tag_id
            next_bid = Builder.stid_bid_map.get((ssid, next_tid), -1)
            head_parent_bp = BasePhrase(event, predicate.pas.event.end, ssid, next_bid, next_tid)
            self.add_children(head_parent_bp, ssid, sentinels=sentinels + [head_bp])
            self.add_compound_phrase_component(head_parent_bp, ssid)
            head_bp.parent = head_parent_bp
            head_parent_bp.children.append(head_bp)

        predicate.head_base_phrase = head_bp
        return head_bp

    def add_compound_phrase_component(self, bp: BasePhrase, ssid: int) -> NoReturn:
        next_tag = Builder.stid_tag_map.get((ssid, bp.tag.tag_id + 1), None)
        if next_tag and "複合辞" in next_tag.features and "補文ト" not in next_tag.features:
            next_tid = bp.tag.tag_id + 1
            next_bid = Builder.stid_bid_map.get((ssid, next_tid), -1)
            parent_bp = BasePhrase(bp.event, next_tag, ssid, next_bid, next_tid)
            self.add_children(parent_bp, ssid, sentinels=[bp])
            self.add_compound_phrase_component(parent_bp, ssid)
            bp.parent = parent_bp
            parent_bp.children.append(bp)

    def add_children(self, parent_bp: BasePhrase, ssid: int, sentinels: List[BasePhrase] = None) -> NoReturn:
        sentinel_tags = {sentinel.tag for sentinel in sentinels} if sentinels else {}
        for child_tag in parent_bp.tag.children:  # type: Tag
            if child_tag in sentinel_tags or "節-主辞" in child_tag.features or "節-区切" in child_tag.features:
                continue
            tid = child_tag.tag_id
            bid = Builder.stid_bid_map.get((ssid, tid), -1)
            child_bp = BasePhrase(parent_bp.event, child_tag, ssid, bid, tid, is_child=True)
            self.add_children(child_bp, ssid, sentinels)
            child_bp.parent = parent_bp
            parent_bp.children.append(child_bp)

    @staticmethod
    def _resolve_duplication(head_bps: List[BasePhrase]) -> NoReturn:
        keys = {head_bp.key[1:] for head_bp in head_bps}  # head_bp.key[0] is the case id.

        def resolver(children: List[BasePhrase]) -> NoReturn:
            for i in reversed(range(len(children))):
                child_bp = children[i]
                if child_bp.omitted_case:
                    continue
                if child_bp.key[1:] in keys:
                    children.pop(i)
                else:
                    resolver(child_bp.children)

        for head in head_bps:
            resolver(head.children)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyknp_eventgraph']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.16,<0.17', 'pyknp>=0.5.0,<0.6.0']

entry_points = \
{'console_scripts': ['evg = pyknp_eventgraph.cli:evg',
                     'evgviz = pyknp_eventgraph.cli:evgviz']}

setup_kwargs = {
    'name': 'pyknp-eventgraph',
    'version': '6.1.7',
    'description': 'A development platform for high-level NLP applications in Japanese',
    'long_description': "# pyknp-eventgraph\n\n**EventGraph** is a development platform for high-level NLP applications in Japanese.\nThe core concept of EventGraph is event, a language information unit that is closely related to predicate-argument structure but more application-oriented.\nEvents are linked to each other based on their syntactic and semantic relations.\n\n## Requirements\n\n- Python 3.6 or later\n- pyknp\n- graphviz\n\n## Installation\n\nTo install pyknp-eventgraph, use `pip`.\n\n```\n$ pip install pyknp-eventgraph\n```\n\n## Quick Tour\n\n### Step 1: Create an EventGraph\n\nAn EventGraph is built on language analysis given in a KNP format.\n\n```python\n# Add imports.\nfrom pyknp import KNP\nfrom pyknp_eventgraph import EventGraph\n\n# Parse a document.\ndocument = ['彼女は海外勤務が長いので、英語がうまいに違いない。', '私はそう確信していた。']\nknp = KNP()\nanalysis = [knp.parse(sentence) for sentence in document]\n\n# Create an EventGraph.\nevg = EventGraph.build(analysis)\nprint(evg)  # <EventGraph, #sentences: 2, #events: 3, #relations: 1>\n```\n\n### Step 2: Extract Information\n\nUsers can obtain various information about language analysis via a simple interface.\n\n#### Step 2.1: Sentence\n\n```python\n# Extract sentences.\nsentences = evg.sentences\nprint(sentences)\n# [\n#   <Sentence, sid: 1, ssid: 0, surf: 彼女は海外勤務が長いので、英語がうまいに違いない。>,\n#   <Sentence, sid: 2, ssid: 1, surf: 私はそう確信していた。>\n# ]\n\n# Convert a sentence into various forms.\nsentence = evg.sentences[0]\nprint(sentence.surf)   # 彼女は海外勤務が長いので、英語がうまいに違いない。\nprint(sentence.mrphs)  # 彼女 は 海外 勤務 が 長い ので 、 英語 が うまい に 違いない 。\nprint(sentence.reps)   # 彼女/かのじょ は/は 海外/かいがい 勤務/きんむ が/が 長い/ながい ので/ので 、/、 英語/えいご が/が 上手い/うまい に/に 違い無い/ちがいない 。/。\n```\n\n#### Step 2.2: Event\n\n```python\n# Extract events.\nevents = evg.events\nprint(events)\n# [\n#   <Event, evid: 0, surf: 海外勤務が長いので、>,\n#   <Event, evid: 1, surf: 彼女は英語がうまいに違いない。>,\n#   <Event, evid: 2, surf: 私はそう確信していた。>\n# ]\n\n# Convert an event into various forms.\nevent = evg.events[0]\nprint(event.surf)              # 海外勤務が長いので、\nprint(event.mrphs)             # 海外 勤務 が 長い ので 、\nprint(event.normalized_mrphs)  # 海外 勤務 が 長い\nprint(event.reps)              # 海外/かいがい 勤務/きんむ が/が 長い/ながい ので/ので 、/、\nprint(event.normalized_reps)   # 海外/かいがい 勤務/きんむ が/が 長い/ながい\nprint(event.content_rep_list)  # ['海外/かいがい', '勤務/きんむ', '長い/ながい']\n\n# Extract an event's PAS.\npas = event.pas\nprint(pas)            # <PAS, predicate: 長い/ながい, arguments: {ガ: 勤務/きんむ}>\nprint(pas.predicate)  # <Predicate, type: 形, surf: 長い>\nprint(pas.arguments)  # defaultdict(<class 'list'>, {'ガ': [<Argument, case: ガ, surf: 勤務が>]})\n\n# Extract an event's features.\nfeatures = event.features\nprint(features)  # <Features, modality: None, tense: 非過去, negation: False, state: 状態述語, complement: False>\n```\n\n#### Step 2.3: Event-to-event Relation\n\n```python\n# Extract event-to-event relations.\nrelations = evg.relations\nprint(relations)  # [<Relation, label: 原因・理由, modifier_evid: 0, head_evid: 1>]\n\n# Take a closer look at an event-to-event relation\nrelation = relations[0]\nprint(relation.label)     # 原因・理由\nprint(relation.surf)      # ので\nprint(relation.modifier)  # <Event, evid: 0, surf: 海外勤務が長いので、>\nprint(relation.head)      # <Event, evid: 1, surf: 彼女は英語がうまいに違いない。>\n```\n\n### Step 3: Seve/Load an EventGraph\n\nUsers can save and load an EventGraph by serializing it as a JSON object.\n\n```python\n# Save an EventGraph as a JSON file.\nevg.save('evg.json')\n\n# Load an EventGraph from a JSON file.\nwith open('evg.json') as f:\n    evg = EventGraph.load(f)\n```\n\n### Step 4: Visualize an EventGraph\n\nUsers can visualize an EventGraph using [graphviz](https://graphviz.org/).\n\n```python\nfrom pyknp_eventgraph import make_image\nmake_image(evg, 'evg.svg')  # Currently, only supports 'svg'.\n```\n\n## Advanced Usage\n\n### Merging modifiers\n\nBy merging a modifier event to the modifiee, users can construct a larger information unit.\n\n```python\nfrom pyknp import KNP\nfrom pyknp_eventgraph import EventGraph\n\ndocument = ['もっととろみが持続する作り方をして欲しい。']\nknp = KNP()\nanalysis = [knp.parse(sentence) for sentence in document]\n\nevg = EventGraph.build(analysis)\nprint(evg)  # <EventGraph, #sentences: 1, #events: 2, #relations: 1>\n\n# Investigate the relation.\nrelation = evg.relations[0]\nprint(relation)           # <Relation, label: 連体修飾, modifier_evid: 0, head_evid: 1>\nprint(relation.modifier)  # <Event, evid: 0, surf: もっととろみが持続する>\nprint(relation.head)      # <Event, evid: 1, surf: 作り方をして欲しい。>\n\n# To merge modifier events, enable `include_modifiers`.\nprint(relation.head.surf)                           # 作り方をして欲しい。\nprint(relation.head.surf_(include_modifiers=True))  # もっととろみが持続する作り方をして欲しい。\n\n# Other formats also support `include_modifiers`.\nprint(relation.head.mrphs_(include_modifiers=True))  # もっと とろみ が 持続 する 作り 方 を して 欲しい 。\nprint(relation.head.normalized_mrphs_(include_modifiers=True))  # もっと とろみ が 持続 する 作り 方 を して 欲しい\n```\n\n### Binary serialization\n\nWhen an EventGraph is serialized in a JSON format, it will lose some functionality, including access to KNP objects and modifier merging.\nTo keep full functionality, use Python's pickle utility for serialization.\n\n```python\n# Save an EventGraph using Python's pickle utility.\nevg.save('evg.pkl', binary=True)\n\n# Load an EventGraph using Python's pickle utility.\nwith open('evg.pkl', 'rb') as f:\n    evg_ = EventGraph.load(f, binary=True)\n```\n\n## CLI\n\n### EventGraph Construction\n\n```\n$ echo '彼女は海外勤務が長いので、英語がうまいに違いない。' | jumanpp | knp -tab | evg -o example-eventgraph.json\n```\n\n### EventGraph Visualization\n\n```\n$ evgviz example-eventgraph.json example-eventgraph.svg\n```\n\n## Documents\n\n[https://pyknp-eventgraph.readthedocs.io/en/latest/](https://pyknp-eventgraph.readthedocs.io/en/latest/)\n\n## Authors\n\n- Kurohashi-Kawahara Lab, Kyoto University.\n- contact@nlp.ist.i.kyoto-u.ac.jp\n",
    'author': 'Hirokazu Kiyomaru',
    'author_email': 'h.kiyomaru@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ku-nlp/pyknp-eventgraph',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from io import open
from typing import List

from pyknp import KNP, BList


def read_knp_result_file(filename: str) -> List[BList]:
    """Read a KNP result file.

    Args:
        filename: A filename.

    Returns:
        A list of :class:`pyknp.knp.blist.BList` objects.
    """
    knp = KNP()
    blists = []
    with open(filename, "rt", encoding="utf-8", errors="replace") as f:
        chunk = ""
        for line in f:
            chunk += line
            if line.strip() == "EOS":
                blists.append(knp.result(chunk))
                chunk = ""
    return blists



from logging import getLogger
from typing import TYPE_CHECKING, List

from pyknp import BList

from pyknp_eventgraph.builder import Builder
from pyknp_eventgraph.component import Component
from pyknp_eventgraph.event import JsonEventBuilder
from pyknp_eventgraph.sentence import JsonSentenceBuilder, Sentence, SentenceBuilder

if TYPE_CHECKING:
    from pyknp_eventgraph.eventgraph import EventGraph


logger = getLogger(__name__)


class Document(Component):
    """A document is a collection of sentences.

    Attributes:
        evg (EventGraph): An EventGraph built on this document.
        sentences (List[Sentence]): A list of sentences in this document.
    """

    def __init__(self, evg: "EventGraph"):
        self.evg: "EventGraph" = evg
        self.sentences: List[Sentence] = []

    def to_dict(self) -> dict:
        """Convert this object into a dictionary."""
        return dict(sentences=[sentence.to_dict() for sentence in self.sentences])

    def to_string(self) -> str:
        """Convert this object into a string."""
        return f"<Document, #sentences: {len(self.sentences)}>"


class DocumentBuilder(Builder):
    def __call__(self, evg: "EventGraph", blists: List[BList]) -> Document:
        document = Document(evg)
        for blist in blists:
            SentenceBuilder()(document, blist)
        evg.document = document
        return document


class JsonDocumentBuilder(Builder):
    def __call__(self, evg: "EventGraph", dump: dict) -> Document:
        document = Document(evg)
        for sentence_dump in dump["sentences"]:
            JsonSentenceBuilder()(document, sentence_dump)
        for event_dump in dump["events"]:
            ssid = event_dump["ssid"]
            JsonEventBuilder()(document.sentences[ssid], event_dump)
        evg.document = document
        return document
